/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <ctime>
#include <cassert>
#include <vector>
#include <sstream>
#include <algorithm>

#include "idlc/c_gen/c_gen_base.h"

namespace {
#include "idlc/c_gen/c_gen_base_cb.h"
}

namespace tidl {

CGeneratorBase::CGeneratorBase(std::shared_ptr<Document> doc)
    : Generator(doc) {
  structures_.clear();
  type_map_ = {
      {"char", "char "}, {"int", "int "}, {"short", "short "},
      {"long", "long long "}, {"bool", "bool "}, {"string", "char *"},
      {"list", "GList *"}, {"float", "float "}, {"double", "double "},
      {"bundle", "bundle *"}, {"void", "void "}
  };
}

std::string CGeneratorBase::Tab(int cnt) {
  std::string str;

  for (int i = 0; i < cnt; ++i) {
    str += "\t";
  }

  return str;
}

std::string CGeneratorBase::NLine(int cnt) {
  std::string str;

  for (int i = 0; i < cnt; ++i) {
    str += "\n";
  }

  return str;
}

std::string CGeneratorBase::ConvertTypeToString(
    ParameterType::Direction direction, const BaseType& type) {
  if (type.IsUserDefinedType()) {
    if (direction == ParameterType::Direction::IN)
      return GetHandlePrefix() + type.ToString() + "_h ";
    else
      return GetHandlePrefix() + type.ToString() + "_h *";
  }

  if (type.ToString() == "array") {
    if (direction == ParameterType::Direction::IN) {
      return GetReturnTypeString(*type.GetMetaType()) + "*";
    } else {
      return GetReturnTypeString(*type.GetMetaType()) + "**";
    }
  }

  if (type.ToString() == "string") {
    if (direction == ParameterType::Direction::IN)
      return "const char *";
    else
      return "char **";
  }

  if (direction == ParameterType::Direction::IN)
    return type_map_[type.ToString()];

  return type_map_[type.ToString()] + "*";
}

std::string CGeneratorBase::GetFullNameFromType(const BaseType& type) {
  std::string str = type.ToString();

  if (type.GetMetaType() != nullptr) {
    str += "_";
    str += GetFullNameFromType(*type.GetMetaType());
  }

  return str;
}

std::string CGeneratorBase::GetParcelParamTypeString(const BaseType& type,
                                                     bool is_pointer) {
  if (type.IsUserDefinedType())
    return GetHandlePrefix() + type.ToString() + "_h ";

  if (type.ToString() == "list" ||
      type.ToString() == "array")
    return GetHandlePrefix() + GetFullNameFromType(type) + "_h ";

  if (type.ToString() == "string")
    return "char *";

  if (is_pointer)
    return type_map_[type.ToString()] + "*";

  return type_map_[type.ToString()];
}

std::string CGeneratorBase::GetReturnTypeString(const BaseType& type) {
  if (type.IsUserDefinedType())
    return GetHandlePrefix() + type.ToString() + "_h ";

  if (type.ToString() == "list" ||
      type.ToString() == "array")
    return GetHandlePrefix() + GetFullNameFromType(type) + "_h ";

  return type_map_[type.ToString()];
}

void CGeneratorBase::AddStructureFromType(const BaseType& type) {
  if (type.GetMetaType() == nullptr)
    return;

  BaseType* t = new BaseType(type);
  assert(t != nullptr);

  std::string type_name = GetFullNameFromType(type);
  Element* elm = new Element(type_name + "s", t, "", __LINE__);
  assert(elm != nullptr);

  Elements* elms = new Elements();
  assert(elms != nullptr);
  elms->Add(elm);

  Structure* st = new Structure(type_name, elms, "", __LINE__);
  assert(st != nullptr);

  if (StructureExist(*st)) {
    delete st;
  } else {
    structures_[type_name] = std::unique_ptr<Structure>(st);
  }

  AddStructureFromType(*type.GetMetaType());
}

std::string CGeneratorBase::SmartIndent(std::string lines) {
  std::stringstream ss(lines);
  std::string result;
  std::string line;
  std::string next_line;
  std::string tab;
  std::string back;
  std::size_t found;
  bool if_statement = false;
  bool continuous = false;
  int tab_size = 0;
  int line_count = 0;

  while (std::getline(ss, next_line, '\n')) {
    line_count++;
    if (line_count == 1) {
      line = Trim(next_line);
      continue;
    }

    tab.clear();
    found = line.find('}');
    if (found != std::string::npos) {
      tab_size--;
    }

    if (line.length() > 0) {
      tab += Tab(tab_size);
    }

    if (continuous && tab_size == 0 &&
        found == std::string::npos) {
      tab += Tab(2);
    }

    if (found == std::string::npos && if_statement) {
      tab_size--;
      if_statement = false;
    }

    continuous = false;
    back = line.back();
    if (back == ",")
      continuous = true;

    if (line.empty() ||
        line.length() == 0 ||
        std::all_of(line.begin(), line.end(), isspace)) {
      std::string n_line = Trim(next_line);
      if (n_line.empty() ||
          n_line.length() == 0 ||
          n_line.find('}') != std::string::npos ||
          std::all_of(n_line.begin(), n_line.end(), isspace)) {
        line = n_line;
        continue;
      } else {
        result += tab + line;
        result += NLine(1);
      }
    } else {
      result += tab + line;
      result += NLine(1);
    }

    found = line.find('{');
    if (found != std::string::npos) {
      tab_size++;
    } else {
      found = line.find("if (");
      if (found != std::string::npos) {
        tab_size++;
        if_statement = true;
      }
    }

    line = Trim(next_line);
  }

  result += line;
  result += NLine(1);

  return result;
}

std::string CGeneratorBase::Trim(const std::string& str) {
  std::size_t first = str.find_first_not_of(" \t\r\n");
  if (first == std::string::npos)
    return str;

  std::size_t last = str.find_last_not_of(" \t\r\n");
  return str.substr(first, (last - first + 1));
}

void CGeneratorBase::GenLicenseDescription(std::ofstream& stream) {
  GenTemplate(CB_COPYRIGHT, stream,
      [&]()->std::string {
        return FULLVER;
      },
      [&]()->std::string {
        std::time_t t = std::time(NULL);
        std::tm* local_time = std::localtime(&t);
        return std::to_string(local_time->tm_year + 1900);
      });
}

void CGeneratorBase::GenIncludeDefaultHeaders(std::ofstream& stream,
                                              bool body) {
  if (body) {
    stream << CB_BODY_HEADER;
  } else {
    stream << CB_HEADER;
  }
}

void CGeneratorBase::GenGNUSourceDefinition(std::ofstream& stream) {
  const char format[] = "#define _GNU_SOURCE\n";

  stream << NLine(1);
  stream << std::string(format);
}

bool CGeneratorBase::StructureExist(const Structure& st) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_STRUCTURE)
      continue;

    const Structure &s = static_cast<const Structure&>(*i);
    if (s.GetID() == st.GetID())
      return true;
  }
  return false;
}

std::string CGeneratorBase::GetParamTypeString(
    ParameterType::Direction direction, const BaseType& type) {
  if (type.IsUserDefinedType() ||
      type.ToString() == "list" ||
      type.ToString() == "array") {
    if (direction == ParameterType::Direction::IN)
      return GetHandlePrefix() + GetFullNameFromType(type) + "_h ";
    else
      return GetHandlePrefix() + GetFullNameFromType(type) + "_h *";
  }

  if (type.ToString() == "string") {
    if (direction == ParameterType::Direction::IN)
      return "const char *";
    else
      return "char **";
  }

  if (direction == ParameterType::Direction::IN)
    return type_map_[type.ToString()];

  return type_map_[type.ToString()] + "*";
}

std::string CGeneratorBase::GetErrorValue(const BaseType& type) {
  if (type.IsUserDefinedType() ||
      type.ToString() == "list" ||
      type.ToString() == "array" ||
      type.ToString() == "bundle" ||
      type.ToString() == "string")
    return "NULL";
  if (type.ToString() == "bool")
    return "false";
  if (type.ToString() == "char")
    return "0";
  return "-1";
}

std::string CGeneratorBase::GetStringFromElementType(const BaseType& type) {
  if (type.IsUserDefinedType())
    return GetHandlePrefix() + type.ToString() + "_h ";
  if (type.ToString() == "array")
    return GetReturnTypeString(*type.GetMetaType()) + "*";

  return type_map_[type.ToString()];
}

std::string CGeneratorBase::GetStructIdWithNamespace(const Structure& st) {
  if (!HasNamespace())
    return st.GetID();
  return GetFileNamespace() + "_" + st.GetID();
}

std::string CGeneratorBase::GetInterfaceIdWithNamespace(const Interface& inf) {
  if (!HasNamespace())
    return inf.GetID();
  return GetFileNamespace() + "_" + inf.GetID();
}

std::string CGeneratorBase::GetHandlePrefix() {
  if (!HasNamespace())
    return "rpc_port_";
  return "rpc_port_" + GetFileNamespace() + "_";
}

}  // namespace tidl
