/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <vector>

#include "idlc/c_gen/c_header_gen_base.h"

namespace {
#include "idlc/c_gen/c_header_gen_base_cb.h"
}

namespace tidl {

CHeaderGeneratorBase::CHeaderGeneratorBase(std::shared_ptr<Document> doc)
    : CGeneratorBase(doc) {}

void CHeaderGeneratorBase::GenPragmaOnce(std::ofstream& stream) {
  const char format[] = "#pragma once\n";

  stream << NLine(1);
  stream << std::string(format);
}

void CHeaderGeneratorBase::GenExplicitLinkageOpen(std::ofstream& stream) {
  stream << CB_EXPLICIT_LINKAGE_OPEN;
}

void CHeaderGeneratorBase::GenExplicitLinkageClose(std::ofstream& stream) {
  stream << CB_EXPLICIT_LINKAGE_CLOSE;
}

void CHeaderGeneratorBase::GenStructures(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() == Block::TYPE_STRUCTURE) {
      const Structure &st = static_cast<const Structure&>(*i);
      GenStructure(stream, st);
      for (auto& j : st.GetElements().GetElms()) {
       auto& t = j->GetType();
       AddStructureFromType(t);
      }
    } else {
      const Interface &inf = static_cast<const Interface&>(*i);
      for (auto& d : inf.GetDeclarations().GetDecls()) {
        for (auto& p : d->GetParameters().GetParams()) {
          if (IsDelegateType(inf, p->GetParameterType().GetBaseType()))
            continue;
          AddStructureFromType(p->GetParameterType().GetBaseType());
        }
      }
    }
  }

  for (auto& p : GetStructures()) {
    const Structure& st = *p.second;
    GenStructure(stream, st);
  }
}

void CHeaderGeneratorBase::GenStructure(std::ofstream& stream,
                                        const Structure& st) {
  GenStructureDeclaration(stream, st);
  GenStructureConstructor(stream, st);
  GenStructureDestructor(stream, st);
  GenStructureCloner(stream, st);
  GenStructureSetter(stream, st);
  GenStructureGetter(stream, st);
  GenStructureIterator(stream, st);
}

void CHeaderGeneratorBase::GenStructureDeclaration(std::ofstream& stream,
                                                   const Structure& st) {
  GenTemplate(CB_STRUCT_DECL, stream,
      [&]()->std::string {
        return GetStructIdWithNamespace(st);
      },
      [&]()->std::string {
        return GetStructIdWithNamespace(st);
      });
}

void CHeaderGeneratorBase::GenStructureConstructor(std::ofstream& stream,
                                                   const Structure& st) {
  GenTemplate(CB_STRUCT_CTOR, stream,
      [&]()->std::string {
        return GetStructIdWithNamespace(st);
      },
      [&]()->std::string {
        return GetStructIdWithNamespace(st);
      });
}

void CHeaderGeneratorBase::GenStructureDestructor(std::ofstream& stream,
                                                  const Structure& st) {
  GenTemplate(CB_STRUCT_DTOR, stream,
      [&]()->std::string {
        return GetStructIdWithNamespace(st);
      },
      [&]()->std::string {
        return GetStructIdWithNamespace(st);
      });
}

void CHeaderGeneratorBase::GenStructureSetter(std::ofstream& stream,
                                              const Structure& st) {
  for (auto& i : st.GetElements().GetElms()) {
    GenTemplate(CB_STRUCT_SETTER, stream,
          [&]()->std::string {
            return GetStructIdWithNamespace(st);
          },
          [&]()->std::string {
            if (i->GetType().ToString() == "list")
              return "add";
            return "set";
          },
          [&]()->std::string {
            return i->GetID();
          },
          [&]()->std::string {
            return GetStructIdWithNamespace(st);
          },
          [&]()->std::string {
            if (i->GetType().IsUserDefinedType())
              return GetParcelParamTypeString(i->GetType());

            if (i->GetType().ToString() == "list") {
              if (i->GetType().GetMetaType()->IsUserDefinedType() ||
                  i->GetType().GetMetaType()->ToString() == "list" ||
                  i->GetType().GetMetaType()->ToString() == "array") {
                return GetParcelParamTypeString(*i->GetType().GetMetaType());
              } else {
                return ConvertTypeToString(ParameterType::Direction::IN,
                    *i->GetType().GetMetaType());
              }
            }

            if (i->GetType().ToString() == "array") {
              return GetStringFromElementType(i->GetType());
            }

            return ConvertTypeToString(ParameterType::Direction::IN,
                i->GetType());
          },
          [&]()->std::string {
            if (i->GetType().ToString() == "array") {
              std::string str;
              str += i->GetID();
              str += ", ";
              str += "int " + i->GetID() + "_size";
              return str;
            }
            return i->GetID();
          });
  }
}

void CHeaderGeneratorBase::GenStructureGetter(std::ofstream& stream,
                                              const Structure& st) {
  for (auto& i : st.GetElements().GetElms()) {
    if (i->GetType().ToString() == "list")
      continue;

    GenTemplate(CB_STRUCT_GETTER, stream,
        [&]()->std::string {
          return GetStructIdWithNamespace(st);
        },
        [&]()->std::string {
          return i->GetID();
        },
        [&]()->std::string {
          return GetStructIdWithNamespace(st);
        },
        [&]()->std::string {
          if (i->GetType().IsUserDefinedType())
            return GetParcelParamTypeString(i->GetType());

          if (i->GetType().ToString() == "array")
              return GetStringFromElementType(i->GetType()) + "*";

          return ConvertTypeToString(ParameterType::Direction::OUT,
              i->GetType());
        },
        [&]()->std::string {
          if (i->GetType().ToString() == "array") {
            std::string str;
            str += i->GetID();
            str += ", ";
            str += "int *" + i->GetID() + "_size";
            return str;
          }
          return i->GetID();
        });
  }
}

void CHeaderGeneratorBase::GenStructureIterator(std::ofstream& stream,
                                                const Structure& st) {
  for (auto& i : st.GetElements().GetElms()) {
    if (i->GetType().ToString() != "list")
      continue;

    GenTemplate(CB_STRUCT_ITERATOR, stream,
        [&]()->std::string {
          return GetStructIdWithNamespace(st);
        },
        [&]()->std::string {
          return i->GetID();
        },
        [&]()->std::string {
          return GetStructIdWithNamespace(st);
        },
        [&]()->std::string {
          if (i->GetType().GetMetaType()->IsUserDefinedType() ||
              i->GetType().GetMetaType()->ToString() == "list" ||
              i->GetType().GetMetaType()->ToString() == "array")
            return GetParcelParamTypeString(*i->GetType().GetMetaType());

          return ConvertTypeToString(ParameterType::Direction::IN,
              *i->GetType().GetMetaType());
        },
        [&]()->std::string {
          return i->GetID();
        });
  }
}

void CHeaderGeneratorBase::GenStructureCloner(std::ofstream& stream,
                                              const Structure& st) {
  GenTemplate(CB_STRUCT_CLONER, stream,
      [&]()->std::string {
        return GetStructIdWithNamespace(st);
      },
      [&]()->std::string {
        return GetStructIdWithNamespace(st);
      },
      [&]()->std::string {
        return GetStructIdWithNamespace(st);
      });
}

}  // namespace tidl
