/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "idlc/c_gen/c_proxy_header_gen.h"

namespace {
#include "idlc/c_gen/c_proxy_header_gen_cb.h"
}

namespace tidl {

CProxyHeaderGen::CProxyHeaderGen(std::shared_ptr<Document> doc)
    : CHeaderGeneratorBase(doc) {}

void CProxyHeaderGen::OnInitGen(std::ofstream& stream) {
  GenLicenseDescription(stream);
  GenPragmaOnce(stream);
  GenIncludeDefaultHeaders(stream, false);
  GenExplicitLinkageOpen(stream);
  GenStructures(stream);
  GenInterfaces(stream);
}

void CProxyHeaderGen::OnFiniGen(std::ofstream& stream) {
  GenExplicitLinkageClose(stream);
}

void CProxyHeaderGen::GenInterfaces(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_INTERFACE)
      continue;

    const Interface &inf = static_cast<const Interface&>(*i);
    GenInterface(stream, inf);
  }
}

void CProxyHeaderGen::GenInterface(std::ofstream& stream,
                                   const Interface& inf) {
  GenInterfaceDeclaration(stream, inf);
  GenInterfaceDelegators(stream, inf);
  GenInterfaceConnect(stream, inf);
  GenInterfaceDisconnect(stream, inf);
  GenInterfaceMethods(stream, inf);
}

void CProxyHeaderGen::GenInterfaceDelegators(std::ofstream& stream,
                                             const Interface& inf) {
  for (auto& i : inf.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() != Declaration::MethodType::DELEGATE)
      continue;
    GenInterfaceDelegator(stream, GetInterfaceIdWithNamespace(inf), *i);
  }
}

void CProxyHeaderGen::GenInterfaceDelegator(std::ofstream& stream,
                                            const std::string& id,
                                            const Declaration& decl) {
  const char handle[] = "typedef struct $$_s *rpc_port_$$_h;";

  stream << NLine(1);
  stream << GenTemplateString(handle,
      [&]()->std::string {
        return id + "_" + decl.GetID();
      },
      [&]()->std::string {
        return id + "_" + decl.GetID();
      });

  stream << NLine(1);
  stream << GenTemplateString(CB_INTERFACE_DELEGATOR,
      [&]()->std::string {
        return GetReturnTypeString(decl.GetType());
      },
      [&]()->std::string {
        return id + "_" + decl.GetID();
      },
      [&]()->std::string {
        std::string str;
        str += "void *user_data";
        for (auto& p : decl.GetParameters().GetParams()) {
          str += ", ";
          str += GetParamTypeString(p->GetParameterType().GetDirection(),
              p->GetParameterType().GetBaseType()) + p->GetID();
        }
        return str;
      });

  stream << ReplaceAll(CB_DELEGATE_CTOR, "##", id + "_" + decl.GetID());
  stream << GenTemplateString(
      ReplaceAll(CB_DELEGATE_DISPOSER, "##", id + "_" + decl.GetID()),
      [&]()->std::string {
        return id;
      });
}

void CProxyHeaderGen::GenInterfaceDeclaration(std::ofstream& stream,
                                              const Interface& inf) {
  stream << SmartIndent(ReplaceAll(CB_INTERFACE_DECL, "##",
      GetInterfaceIdWithNamespace(inf)));
}

void CProxyHeaderGen::GenInterfaceMethods(std::ofstream& stream,
                                          const Interface& inf) {
  for (auto& i : inf.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
      continue;
    stream << GenTemplateString(
        ReplaceAll(CB_INTERFACE_METHODS, "##",
            GetInterfaceIdWithNamespace(inf)),
        [&]()->std::string {
          return GetReturnTypeString(i->GetType());
        },
        [&]()->std::string {
          return i->GetID();
        },
        [&]()->std::string {
          std::string str;
          for (auto& p : i->GetParameters().GetParams()) {
            str += ", ";
            if (IsDelegateType(inf, p->GetParameterType().GetBaseType())) {
              str += "rpc_port_" + GetInterfaceIdWithNamespace(inf) + "_" +
                  p->GetParameterType().GetBaseType().ToString() +
                  "_h " + p->GetID();
            } else {
              str += GetParamTypeString(p->GetParameterType().GetDirection(),
                  p->GetParameterType().GetBaseType()) + p->GetID();
            }
          }
          return str;
        });
  }
}

void CProxyHeaderGen::GenInterfaceConnect(std::ofstream& stream,
                                          const Interface& inf) {
  stream << ReplaceAll(CB_INTERFACE_CONNECT, "##",
      GetInterfaceIdWithNamespace(inf));
}

void CProxyHeaderGen::GenInterfaceDisconnect(std::ofstream& stream,
                                             const Interface& inf) {
  stream << ReplaceAll(CB_INTERFACE_DISCONNECT, "##",
      GetInterfaceIdWithNamespace(inf));
}

}  // namespace tidl
