/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "idlc/cs_gen/cs_stub_gen.h"

namespace {
#include "idlc/cs_gen/cs_stub_gen_cb.h"
}

namespace tidl {

CsStubGen::CsStubGen(std::shared_ptr<Document> doc)
    : CsGeneratorBase(doc) {}

void CsStubGen::OnInitGen(std::ofstream& stream) {
  GenCopyright(stream);
  stream << "using System;" << NLine(1)
         << "using System.Collections.Generic;" << NLine(1)
         << "using Tizen.Applications;" << NLine(1)
         << "using Tizen.Applications.RPCPort;" << NLine(1);

  GenNamespace(stream);
}

void CsStubGen::OnFiniGen(std::ofstream& stream) {}

void CsStubGen::GenNamespace(std::ofstream& stream) {
  stream << "namespace RPCPort" << NLine(1);
  GenBrace(stream, 0, [&]() {
    stream << "namespace " << GetFileNamespace() << NLine(1);
    GenBrace(stream, 0, [&]() {
      GenStructures(stream);
      stream << Tab(1) << "namespace Stub" << NLine(1);
      GenBrace(stream, TAB_SIZE, [&]() {
        GenInterfaces(stream);
      });
    });
  });
}

void CsStubGen::GenInterfaces(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_INTERFACE)
      continue;
    Interface& iface = static_cast<Interface&>(*i);
    GenInterface(stream, iface);
    stream << std::endl;
  }
}

void CsStubGen::GenInterface(std::ofstream& stream, const Interface& iface) {
  stream << Tab(2) << "public sealed class " << iface.GetID()
         << " : StubBase" << NLine(1);
  GenBrace(stream, TAB_SIZE * 2, [&]() {
    stream << CB_DATA_MEMBERS;
    GenServiceBase(stream, iface);
    GenCallbacks(stream, iface, false);
    GenDelegateId(stream, iface);
    GenMethodId(stream, iface);
    GenSerializer(stream);
    GenListSerializer(stream);
    GenReceivedEvent(stream, iface);
    GenConnectedEvent(stream);
    GenDisconnectedEvent(stream);
    GenCtor(stream, iface);
    GenCommonMethods(stream);
    GenDisposable(stream, iface);
  });
}

void CsStubGen::GenServiceBase(std::ofstream& stream, const Interface& iface) {
  stream << CB_SERVICE_BASE_FRONT;
  GenDeclarations(stream, iface.GetDeclarations());
  stream << NLine(1);
  stream << AddIndent(TAB_SIZE * 3, "}\n");
}

void CsStubGen::GenDeclarations(std::ofstream& stream,
                                const Declarations& decls) {
  for (auto& i : decls.GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
      continue;
    if (!i->GetComments().empty())
      stream << NLine(1) << AddIndent(TAB_SIZE * 4, i->GetComments());
    stream << Tab(4) << "public abstract ";
    GenDeclaration(stream, *i);
    stream << NLine(1);
  }
}

void CsStubGen::GenReceivedEvent(std::ofstream& stream,
    const Interface& iface) {
  stream << CB_ON_RECEIVED_EVENT_FRONT;
  for (auto& i : iface.GetDeclarations().GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
      continue;
    stream << Tab(7) << "case MethodId." << i->GetID() << ":" << NLine(1);
    GenBrace(stream, TAB_SIZE * 8, [&]() {
      GenInvocation(stream, *i);
      stream << Tab(9) << "break;" << NLine(1);
    });
  }
  stream << CB_ON_RECEIVED_EVENT_BACK;
}

void CsStubGen::GenInvocation(std::ofstream& stream, const Declaration& decl) {
  int cnt = 1;

  // Deserialize
  for (auto& i : decl.GetParameters().GetParams()) {
    if (i->GetParameterType().GetDirection() == ParameterType::Direction::OUT) {
      cnt++;
      continue;
    }

    std::string v = "param" + std::to_string(cnt);
    std::string c = ConvertTypeToDeserializer(
        i->GetParameterType().GetBaseType(), v, "p");
    stream << AddIndent(TAB_SIZE * 9, c);
    cnt++;
  }

  // Invoke
  cnt = 1;
  std::string m;
  bool hasRet = false;

  if (decl.GetType().ToString() != "void") {
    m += "var retVal = ";
    hasRet = true;
  }

  m += "b." + decl.GetID() + "(";
  for (auto& i : decl.GetParameters().GetParams()) {
    if (cnt != 1) {
      m += ", ";
    }

    std::string v = "param" + std::to_string(cnt);
    auto& pt = i->GetParameterType();
    if (pt.GetDirection() == ParameterType::Direction::OUT) {
      m += "out " + ConvertTypeToString(pt.GetBaseType()) + " ";
    } else if (pt.GetDirection() == ParameterType::Direction::REF) {
      m += "ref ";
    }
    m += v;
    cnt++;
  }

  m += ");\n";
  stream << AddIndent(TAB_SIZE * 9, m);

  // Serialize
  if (decl.GetMethodType() == Declaration::MethodType::ASYNC)
    return;

  cnt = 0;
  m = "result.WriteInt((int)MethodId.__Result);\n";
  for (auto& i : decl.GetParameters().GetParams()) {
    auto& pt = i->GetParameterType();
    cnt++;
    if (pt.GetDirection() == ParameterType::Direction::IN)
      continue;
    m += ConvertTypeToSerializer(pt.GetBaseType(),
                                "param" + std::to_string(cnt), "result");
  }

  if (hasRet) {
    m += ConvertTypeToSerializer(decl.GetType(), "retVal", "result");
  }

  m += "result.Send(port);\n";
  stream << AddIndent(TAB_SIZE * 9, m);
}

void CsStubGen::GenConnectedEvent(std::ofstream& stream) {
  stream << CB_ON_CONNECTED_EVENT;
}

void CsStubGen::GenDisconnectedEvent(std::ofstream& stream) {
  stream << CB_ON_DISCONNECTED_EVENT;
}

void CsStubGen::GenCtor(std::ofstream& stream, const Interface& iface) {
  GenTemplate(CB_CTOR_FRONT, stream,
    [&]()->std::string {
      return iface.GetID();
    },
    [&]()->std::string {
      return iface.GetID();
    });

  for (auto& i : iface.GetAttributes().GetAttrs()) {
    if (i->GetKey() == "privilege") {
      stream << Tab(4) << "AddPrivilege(\""
             << i->GetValue() << "\");" << NLine(1);
    } else if (i->GetKey() == "trusted" && i->GetValue() == "true") {
      stream << Tab(4) << "SetTrusted("
             << i->GetValue() << ");" << NLine(1);
    }
  }
  stream << Tab(3) << "}" << NLine(1);
}

void CsStubGen::GenCommonMethods(std::ofstream& stream) {
  stream << CB_COMMON_METHODS;
}

void CsStubGen::GenDisposable(std::ofstream& stream, const Interface& iface) {
  stream << CB_DISPOSABLE;
}

}  // namespace tidl
