/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <gtest/gtest.h>

#include <iostream>

#include "idlc/type.h"

class TokenTest : public testing::Test {
 public:
  tidl::Token* testToken;

  virtual void SetUp() {
    testToken = new tidl::Token("TestToken", "Test Token");
  }
  virtual void TearDown() {
    delete testToken;
  }
};

TEST_F(TokenTest, Token_Constructor) {
  tidl::Token* token = new tidl::Token("TestToken", "");
  EXPECT_NE(token, nullptr);
  delete token;
}

TEST_F(TokenTest, Token_ToString) {
  EXPECT_EQ(testToken->ToString(), "TestToken");
}

TEST_F(TokenTest, Token_GetComments) {
  EXPECT_EQ(testToken->GetComments(), "Test Token");
}

class BaseTypeTest : public testing::Test {
 public:
  virtual void SetUp() {}
  virtual void TearDown() {}
};

TEST_F(BaseTypeTest, BaseType_Constructor) {
  tidl::BaseType* baseType = new tidl::BaseType("BaseType", "");
  EXPECT_NE(baseType, nullptr);
  delete baseType;
}

TEST_F(BaseTypeTest, BaseType_SetMetaType) {
  tidl::BaseType* customType = new tidl::BaseType("CustomType", "", true);
  customType->SetMetaType(new tidl::BaseType("int", ""));
  EXPECT_EQ(customType->GetMetaType()->ToString(), "int");
  delete customType;
}

TEST_F(BaseTypeTest, BaseType_GetMetaType) {
  tidl::BaseType* customType = new tidl::BaseType("CustomType", "", true);
  customType->SetMetaType(new tidl::BaseType("string", ""));
  EXPECT_EQ(customType->GetMetaType()->ToString(), "string");
  delete customType;
}

TEST_F(BaseTypeTest, BaseType_GetFullName) {
  tidl::BaseType* customType = new tidl::BaseType("CustomType", "", true);
  customType->SetMetaType(new tidl::BaseType("string", ""));
  EXPECT_EQ(customType->GetFullName(), "CustomType<string>");
  delete customType;
}

TEST_F(BaseTypeTest, BaseType_IsUserDefinedType) {
  tidl::BaseType* testType = new tidl::BaseType("TestType", "", true);
  EXPECT_EQ(testType->IsUserDefinedType(), true);
  delete testType;
}

TEST_F(BaseTypeTest, BaseType_GetComments) {
  std::string comments = "Test BaseType";
  tidl::BaseType* testType = new tidl::BaseType("TestType", comments, true);
  EXPECT_NE(testType, nullptr);
  EXPECT_EQ(testType->GetComments(), comments);
  delete testType;
}

class ParameterTypeTest : public testing::Test {
 public:
  virtual void SetUp() {}
  virtual void TearDown() {}
};

TEST_F(ParameterTypeTest, ParameterType_Constructor) {
  tidl::ParameterType *parameterType = new tidl::ParameterType(
      new tidl::BaseType("int", ""));
  EXPECT_NE(parameterType, nullptr);
  delete parameterType;
}

TEST_F(ParameterTypeTest, ParameterType_Constructor_With_Direction) {
  tidl::ParameterType *parameterType = new tidl::ParameterType(
      new tidl::BaseType("int", ""), "in");
  EXPECT_NE(parameterType, nullptr);
  delete parameterType;
}

TEST_F(ParameterTypeTest, ParameterType_GetDirection) {
  tidl::ParameterType *parameterType = new tidl::ParameterType(
      new tidl::BaseType("int", ""), "out");
  EXPECT_NE(parameterType, nullptr);
  EXPECT_EQ(parameterType->GetDirection(), tidl::ParameterType::Direction::OUT);
  delete parameterType;
}

TEST_F(ParameterTypeTest, ParameterType_GetBaseType) {
  tidl::ParameterType *parameterType = new tidl::ParameterType(
      new tidl::BaseType("string", ""), "ref");
  EXPECT_NE(parameterType, nullptr);
  EXPECT_EQ(parameterType->GetBaseType().ToString(),
      "string");
  delete parameterType;
}
